"构建数据包方便获取"

class PacketIDS:
    "数据包id"
    Text = 9
    PlayerList = 63
    CommandOutput = 79


class SubPacket_CmdOutputMsg:
    """命令输出消息子包构建"""
    Success: bool
    Message: str
    Parameters: list[str]

    def __init__(self, pkt: dict):
        self.Success = pkt["Success"]
        self.Parameters = pkt["Parameters"]
        self.Message = pkt["Message"]


class SubPacket_CmdOrigin:
    "命令来源子包构建"
    Origin: int
    UUID: str
    RequestID: str
    PlayerUniqueID: int

    def __init__(self, pkt: dict):
        self.Origin = pkt["Origin"]
        self.UUID = pkt["UUID"]
        self.RequestID = pkt["RequestID"]
        self.PlayerUniqueID = pkt["PlayerUniqueID"]


class Packet_CommandOutput:
    "命令输出包构建"
    CommandOrigin: SubPacket_CmdOrigin
    OutputType: int
    SuccessCount: int
    OutputMessages: list[SubPacket_CmdOutputMsg]
    as_dict: dict

    def __init__(self, pkt: dict):
        self.as_dict = pkt
        self.CommandOrigin = SubPacket_CmdOrigin(pkt["CommandOrigin"])
        self.OutputMessages = [
            SubPacket_CmdOutputMsg(imsg) for imsg in pkt["OutputMessages"]
        ]
        self.SuccessCount = pkt["SuccessCount"]
        self.OutputType = pkt["OutputType"]



"""
提供了一些实用方法的类

Classes:
- Builtins: 一个提供了线程、JSON操作和文件操作的实用方法的类。

Methods:
- ThreadExit: 用于线程终止的SystemExit的子类。
- ClassicThread: 简化ToolDelta子线程创建的threading.Thread的子类。
- TMPJson: 提供了加载、卸载、读取和写入JSON文件到缓存区的方法的类。
- SimpleJsonDataReader: 提供了安全的JSON文件操作方法的类。
- get_threads_list: 返回使用createThread创建的所有线程的列表。
- SimpleFmt: 使用字典中的值替换字符串中的占位符的静态方法。
"""

from io import TextIOWrapper
import os
import time
import copy
from typing import Any, Callable, Dict, List, Tuple, Optional
import ctypes
import threading
import traceback
import json as rjson
import ujson as json
from .color_print import Print

event_pool = {"tmpjson_save": threading.Event()}
event_flags_pool = {"tmpjson_save": True}
threads_list: list["Builtins.createThread"] = []


class Builtins:
    """提供了一些实用方法的类"""
    class ThreadExit(SystemExit):
        """线程退出."""

    class ClassicThread(threading.Thread):
        """简化ToolDelta子线程创建的threading.Thread的子类."""

        def __init__(self, func: Callable, args: tuple = (), usage="", **kwargs):
            """新建一个ToolDelta子线程

            Args:
                func (Callable): 线程方法
                args (tuple, optional):方法的参数项
                usage (str, optional): 线程的用途说明
                kwargs (dict, optional): 方法的关键词参数项
            """
            super().__init__(target=func)
            self.func = func
            self.daemon = True
            self.all_args = [args, kwargs]
            self.usage = usage
            self.start()
            self.stopping = False
            self._thread_id = None

        def run(self) -> None:
            """线程运行方法"""
            threads_list.append(self)
            try:
                self.func(*self.all_args[0], **self.all_args[1])
            except (Builtins.ThreadExit, SystemExit):
                pass
            except Exception:
                Print.print_err(
                    f"线程 {self.usage or self.func.__name__} 出错:\n" + traceback.format_exc())
            finally:
                threads_list.remove(self)

        def get_id(self) -> int:
            """获取线程的ID

            Raises:
                RuntimeError: 线程ID未知

            Returns:
                int: 线程ID
            """
            if self._thread_id is None:
                for thread in threading.enumerate():
                    if thread is self:
                        self._thread_id = thread.ident
                        break
            if self._thread_id is None:
                raise RuntimeError("Could not determine the thread's ID")
            return self._thread_id

        def stop(self) -> None:
            """终止线程"""
            self.stopping = True
            self._thread_id = self.ident
            thread_id = self.get_id()
            res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
                thread_id, ctypes.py_object(SystemExit))
            if res > 1:
                ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)
                Print.print_err("§c终止线程失败")

    createThread = ClassicThread

    class TMPJson:
        """提供了加载、卸载、读取和写入JSON文件到缓存区的方法的类."""
        @staticmethod
        def loadPathJson(path: str, needFileExists: bool = True) -> None:
            """
            将json文件从磁盘加载到缓存区, 以便快速读写.
            在缓存文件已加载的情况下, 再使用一次该方法不会有任何作用.

            Args:
                path (str): 作为文件的磁盘内路径的同时也会作为在缓存区的虚拟路径
                needFileExists (bool, optional): 默认为 True, 为 False 时, 若文件路径不存在, 就会自动创建一个文件, 且默认值为null

            Raises:
                err: 文件不存在时
            """
            if path in jsonPathTmp:
                return
            try:
                with open(path, 'r', encoding='utf-8') as file:
                    js = Builtins.SimpleJsonDataReader.SafeJsonLoad(file)
            except FileNotFoundError as err:
                if not needFileExists:
                    js = None
                else:
                    raise err from None
            jsonPathTmp[path] = [False, js]

        @staticmethod
        def unloadPathJson(path: str) -> bool:
            """
            将json文件从缓存区卸载(保存内容到磁盘), 之后不能再在缓存区对这个文件进行读写.
            在缓存文件已卸载的情况下, 再使用一次该方法不会有任何作用, 但是可以通过其返回的值来知道存盘有没有成功.

            Args:
                path (str): 文件的虚拟路径

            Returns:
                bool: 存盘是否成功
            """
            if jsonPathTmp.get(path) is not None:
                isChanged, dat = jsonPathTmp[path]
                if isChanged:
                    with open(path, 'w', encoding='utf-8') as file:
                        Builtins.SimpleJsonDataReader.SafeJsonDump(dat, file)
                del jsonPathTmp[path]
                return True
            return False

        @staticmethod
        def read(path: str) -> list[Any] | dict[Any, Any] | Any:
            """对缓存区的该虚拟路径的文件进行读操作, 返回一个深拷贝的JSON对象

            Args:
                path (str): 文件的虚拟路径

            Raises:
                Exception: json路径未初始化, 不能进行读取和写入操作

            Returns:
                list[Any] | dict[Any, Any] | Any: 该虚拟路径的JSON
            """
            if path in jsonPathTmp:
                val = jsonPathTmp.get(path)
                if val is not None:
                    val = val[1]
                    if isinstance(val, (list, dict)):
                        val = copy.deepcopy(val)
                    return val
            raise ValueError("json路径未初始化, 不能进行读取和写入操作: " + path)

        @staticmethod
        def get(path: str) -> None:
            """
            直接获取缓存区的该虚拟路径的JSON, 不使用copy
            WARNING: 如果你不知道有什么后果, 请老老实实使用`read(...)`而不是`get(...)`!

            Args:
                path (str): 文件的虚拟路径
            """
            if path in jsonPathTmp:
                val = jsonPathTmp.get(path)
                if val is not None:
                    val = val[1]
                    return val
            raise ValueError("json路径未初始化, 不能进行读取和写入操作: " + path)

        @staticmethod
        def write(path: str, obj: Any) -> None:
            """
            对缓存区的该虚拟路径的文件进行写操作, 这将会覆盖之前的内容

            Args:
                path (str): 文件的虚拟路径
                obj (Any): 任何合法的JSON类型 例如 dict/list/str/bool/int/float
            """
            if path in jsonPathTmp:
                jsonPathTmp[path] = [True, obj]
            else:
                raise ValueError("json路径未初始化, 不能进行读取和写入操作: " + path)

        @staticmethod
        def read_as_tmp(path: str, needFileExists: bool = True, timeout: int = 60) -> Any:
            """读取json文件并将其从磁盘加载到缓存区, 以便一段时间内能快速读写.

            Args:
                path (str): 作为文件的磁盘内路径的同时也会作为在缓存区的虚拟路径
                needFileExists (bool, optional): 默认为 True, 为 False 时, 若文件路径不存在, 就会自动创建一个文件, 且写入默认值null
                timeout (int, optional): 多久没有再进行读取操作时卸载缓存

            Returns:
                Any: 该虚拟路径的JSON
            """
            if path not in jsonUnloadPathTmp and not path in jsonPathTmp:
                jsonUnloadPathTmp[path] = timeout + int(time.time())
                Builtins.TMPJson.loadPathJson(path, needFileExists)
            return Builtins.TMPJson.read(path)

        @staticmethod
        def write_as_tmp(path: str, obj: Any, needFileExists: bool = True, timeout: int = 60) -> None:
            """写入json文件并将其从磁盘加载到缓存区, 以便一段时间内能快速读写.

            Args:
                path (str): 作为文件的磁盘内路径的同时也会作为在缓存区的虚拟路径
                obj (Any): 任何合法的JSON类型 例如 dict/list/str/bool/int/float
                needFileExists (bool, optional): 默认为 True, 为 False 时, 若文件路径不存在, 就会自动创建一个文件, 且写入默认值null
                timeout (int, optional): 多久没有再进行读取操作时卸载缓存
            """
            if path not in jsonUnloadPathTmp and not path in jsonPathTmp:
                jsonUnloadPathTmp[path] = timeout + int(time.time())
                Builtins.TMPJson.loadPathJson(path, needFileExists)
            Builtins.TMPJson.write(path, obj)

        @staticmethod
        def cancel_change(path: str) -> None:
            """取消缓存json所做的更改, 非必要情况请勿调用, 你不知道什么时候会自动保存所做更改"""
            jsonPathTmp[path][0] = False

        @staticmethod
        def get_tmps() -> Dict:
            """不要调用!"""
            return jsonPathTmp.copy()

    class SimpleJsonDataReader:
        """提供了安全的JSON文件操作方法的类."""
        @staticmethod
        def SafeJsonDump(obj: str | dict | list, fp: TextIOWrapper) -> None:
            """将一个json对象写入一个文件, 会自动关闭文件读写接口.

            Args:
                obj (str | dict | list): JSON对象
                fp (_type_): open(...)打开的文件读写口 或 文件路径
            """
            if isinstance(fp, str):
                with open(fp, "w", encoding="utf-8") as file:
                    file.write(json.dumps(obj, indent=4, ensure_ascii=False))
            else:
                with fp:
                    fp.write(json.dumps(obj, indent=4, ensure_ascii=False))

        @staticmethod
        def SafeJsonLoad(fp: TextIOWrapper) -> dict | list:
            """从一个文件读取json对象, 会自动关闭文件读写接口.

            Args:
                fp (TextIOWrapper): open(...)打开的文件读写口

            Returns:
                dict | list: JSON对象
            """
            if isinstance(fp, str):
                with open(fp, "r", encoding="utf-8") as file:
                    return json.loads(file.read())
            with fp as file:
                return json.loads(file.read())

        class DataReadError(json.JSONDecodeError):
            """读取数据时发生错误"""

        @staticmethod
        def readFileFrom(plugin_name: str, file: str, default: dict | None = None) -> dict | list:
            """从插件数据文件夹读取一个json文件, 会自动创建文件夹和文件.

            Args:
                plugin_name (str): 插件名
                file (str): 文件名
                default (dict, optional): 默认值, 若文件不存在则会写入这个默认值

            Raises:
                Builtins.SimpleJsonDataReader.DataReadError: 读取数据时发生错误
                err: 读取文件路径时发生错误

            Returns:
                dict | list: JSON对象
            """
            if file.endswith(".json"):
                file = file[:-5]
            filepath = os.path.join("插件数据文件", plugin_name, f"{file}.json")
            os.makedirs(os.path.join("插件数据文件", plugin_name), exist_ok=True)
            try:
                if default is not None and not os.path.isfile(filepath):
                    with open(filepath, "w", encoding="utf-8") as f:
                        Builtins.SimpleJsonDataReader.SafeJsonDump(default, f)
                    return default
                with open(filepath, "r", encoding="utf-8") as f:
                    res = Builtins.SimpleJsonDataReader.SafeJsonLoad(f)
                return res
            except rjson.JSONDecodeError as err:
                # 判断是否有msg.doc.pos属性
                raise Builtins.SimpleJsonDataReader.DataReadError(
                    err.msg, err.doc, err.pos
                )
            except Exception as err:
                Print.print_err(f"读文件路径 {filepath} 发生错误")
                raise err

        @staticmethod
        def writeFileTo(plugin_name: str, file: str, obj: str | dict[Any, Any] | list[Any]) -> None:
            """将一个json对象写入插件数据文件夹, 会自动创建文件夹和文件.

            Args:
                plugin_name (str): 插件名
                file (str): 文件名
                obj (str | dict[Any, Any] | list[Any]): JSON对象
            """
            os.makedirs(f"插件数据文件/{plugin_name}", exist_ok=True)
            with open(f"插件数据文件/{plugin_name}/{file}.json", "w", encoding="utf-8") as f:
                Builtins.SimpleJsonDataReader.SafeJsonDump(obj, f)

    @staticmethod
    def get_threads_list() -> list["Builtins.createThread"]:
        "返回使用 createThread 创建的全线程列表."
        return threads_list

    @staticmethod
    def SimpleFmt(kw: Dict[str, Any], *args: str) -> str:
        """
        快速将字符串内按照给出的dict的键值对替换掉内容.

        参数:
            kw: Dict[str, Any], 键值对应替换的内容
            *args: str, 需要被替换的字符串

        示例:
            >>> my_color = "red"; my_item = "apple"
            >>> kw = {"[颜色]": my_color, "[物品]": my_item}
            >>> Builtins.SimpleFmt(kw, "I like [颜色] [物品].")
            I like red apple.
        """
        __sub = args[0]
        for k, v in kw.items():
            if k in __sub:
                __sub = __sub.replace(k, str(v))
        return __sub

    @staticmethod
    def simpleAssert(cond: Any, exc: Any) -> None:
        """
        相当于 assert cond, 但是可以自定义引发的异常的类型
        """
        if not cond:
            raise exc

    @staticmethod
    def thread_func(func: Any) -> Any:
        """
        在事件方法可能执行较久会造成堵塞时使用, 方便快捷地创建一个新线程, 例如:

        @Builtins.run_as_new_thread
        def on_inject(self):
            ...
        """

        def thread_fun(*args: Tuple, **kwargs: Any) -> None:
            Builtins.createThread(func, usage="简易线程方法:" +
                                  func.__name__, args=args, **kwargs)

        return thread_fun

    run_as_new_thread = thread_func
    new_thread = thread_func

    @staticmethod
    def try_int(arg: Any) -> Optional[int]:
        """尝试将提供的参数化为int类型并返回, 否则返回None"""
        try:
            return int(arg)
        except Exception:
            return None

    @staticmethod
    def add_in_dialogue_player(player: str) -> None:
        """
        使玩家进入聊天栏对话模式, 可防止其在对话时继续触发另一个会话线程

        参数:
            player: str, 玩家名
        """
        if player not in in_dialogue_list:
            in_dialogue_list.append(player)
        else:
            raise ValueError("Already in a dialogue!")

    @staticmethod
    def remove_in_dialogue_player(player: str) -> None:
        """
        使玩家离开聊天栏对话模式

        参数:
            player: str, 玩家名
        """
        if player not in in_dialogue_list:
            return
        in_dialogue_list.remove(player)

    @staticmethod
    def player_in_dialogue(player: str) -> bool:
        """
        检测玩家是否处在聊天栏对话模式中.

        参数:
            player: str, 玩家名
        返回:
            bool, 检测结果
        """
        return player in in_dialogue_list

    @staticmethod
    def create_dialogue_threading(
        player: str,
        func: Any,
        exc_cb: Optional[Any] = None,
        args: Tuple = (),
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        创建一个玩家与聊天栏交互的线程,
        线程启动时玩家会自动进入聊天栏对话模式
        线程结束后该玩家会自动退出聊天栏对话模式
        可以用来防止玩家多开聊天栏对话或菜单线程

        参数:
            player: str, 玩家名
            func: function, 线程方法
            exc_cb: function, 若玩家已处于一个对话中, 则向方法exc_cb传参并调用它: player(玩家名)
            args: tuple, 线程方法的参数组
            kwargs: dict, 线程方法的关键词参数组
        """
        if kwargs is None:
            kwargs = {}
        Builtins.createThread(
            _dialogue_thread_run, args=(player, func, exc_cb, args, kwargs)
        )

    @staticmethod
    def fuzzy_match(lst: List[str], sub: str) -> List[str]:
        """
        模糊匹配列表内的字符串, 可以用在诸如模糊匹配玩家名的用途

        参数:
            lst: list, 字符串列表
            sub: str, 需要匹配的字符串
        返回:
            list, 匹配结果
        """
        res = []
        for i in lst:
            if sub in i:
                res.append(i)
        return res

    class ArgsReplacement:
        """用于替换字符串中的占位符的类"""

        def __init__(self, kw: Dict[str, Any]):
            self.kw = kw

        def replaceTo(self, __sub: str) -> str:
            """

            Args:
                __sub (str): 需要替换的字符串

            Returns:
                str: 替换后的字符串
            """
            for k, v in self.kw.items():
                if k in __sub:
                    __sub = __sub.replace(k, str(v))
            return __sub


def safe_close() -> None:
    """安全关闭"""
    event_pool["tmpjson_save"].set()
    event_flags_pool["tmpjson_save"] = False


def _tmpjson_save_thread():
    evt = event_pool["tmpjson_save"]
    secs = 0
    while 1:
        evt.wait(2)
        secs += 2
        if secs >= 60:
            secs = 0
            for k, (isChanged, dat) in jsonPathTmp.copy().items():
                if isChanged:
                    Builtins.SimpleJsonDataReader.SafeJsonDump(dat, k)
                    jsonPathTmp[k][0] = False
        for k, v in jsonUnloadPathTmp.copy().items():
            if time.time() - v > 0:
                Builtins.TMPJson.unloadPathJson(k)
                del jsonUnloadPathTmp[k]
        if not event_flags_pool["tmpjson_save"]:
            return


def tmpjson_save_thread() -> None:
    """JSON缓存文件定时保存"""
    Builtins.createThread(_tmpjson_save_thread, usage="JSON缓存文件定时保存")


def _dialogue_thread_run(player, func, exc_cb, args, kwargs):
    if not Builtins.player_in_dialogue(player):
        Builtins.add_in_dialogue_player(player)
    else:
        if exc_cb is not None:
            exc_cb(player)
        return
    try:
        func(*args, **kwargs)
    except Exception:
        Print.print_err(f"玩家{player}的会话线程 出现问题:")
        Print.print_err(traceback.format_exc())
    Builtins.remove_in_dialogue_player(player)


jsonPathTmp = {}
in_dialogue_list = []
jsonUnloadPathTmp = {}



"启动选项"
import os
import signal
from .color_print import Print
from .starter import start_tool_delta
from .plugin_manager import plugin_manager
from .plugin_market import market
from .sys_args import sys_args_to_dict


def signal_handler(*_) -> None:
    """排除信号中断"""
    return Print.print_war("ToolDelta 已忽略信号中断")


signal.signal(signal.SIGINT, signal_handler)


def client_title() -> None:
    "选择启动模式"
    launch_mode = sys_args_to_dict()
    if launch_mode.get("l"):
        if not isinstance(launch_mode["l"], str):
            raise ValueError("启动模式参数不合法")
        r = launch_mode["l"]
    else:
        Print.clean_print("§b请选择启动模式(使用启动参数 -l <启动模式> 可以跳过该页面):")
        Print.clean_print("1 - 启动 ToolDelta")
        Print.clean_print("2 - 打开 ToolDelta 插件管理器")
        Print.clean_print("3 - 打开 ToolDelta 插件市场")
        r = input("请选择:").strip()
    match r:
        case "1":
            start_tool_delta()
        case "2":
            plugin_manager.manage_plugins()
        case "3":
            market.enter_plugin_market()
        case _:
            Print.clean_print("§c不合法的模式: " + r)
    os._exit(0)



"还原游戏常见字符串"
import os
import re
import tarfile
import gzip
from importlib import util
import threading
from glob import glob
from typing import Dict
import requests
import ujson as json
from .urlmethod import download_file_singlethreaded
from .color_print import Print


class GameTextsLoader:
    "还原游戏常见字符串"

    def __init__(self) -> None:
        "初始化"
        self.base_path = os.path.join(os.getcwd(), "插件数据文件", "game_texts")
        self.check_initial_run()
        self.start_auto_update_thread()
        self.game_texts_data: Dict[str, str] = self.load_data()

    @staticmethod
    def get_latest_version() -> str:
        """获取最新版本号

        Returns:
            str: 版本号
        """
        result = re.match(
            r"(\d+\.\d+\.\d+)",
            requests.get(
                "https://api.github.com/repos/ToolDelta/GameText/releases/latest", timeout=5
            ).json()["tag_name"],
        )
        if not isinstance(result, type(None)):
            return result.group()
        raise ValueError("无法获取最新版本号")

    def check_initial_run(self) -> None:
        "检查初始运行"
        version_file_path: str = os.path.join(self.base_path, "version")
        if not os.path.exists(version_file_path):
            latest_version: str = self.get_latest_version()
            with open(version_file_path, "w", encoding="utf-8") as f:
                f.write(latest_version)
            self.download_and_extract(latest_version)

    def start_auto_update_thread(self) -> None:
        "启用自动更新线程"
        threading.Timer(24 * 60 * 60, self.auto_update).start()

    def auto_update(self) -> None:
        "自动更新"
        version_file_path: str = os.path.join(self.base_path, "version")
        with open(version_file_path, "r", encoding="utf-8") as f:
            version: str = f.read()
        latest_version: str = self.get_latest_version()
        if version != latest_version:
            self.download_and_extract(latest_version)
        threading.Timer(24 * 60 * 60, self.auto_update).start()

    def download_and_extract(self, version) -> None:
        "下载并解压"
        packets_url: str = (
            f"https://hub.gitmirror.com/?q=https://github.com/ToolDelta/GameText/releases/download/{version}/ToolDelta_Game_Texts.tar.gz"
        )
        archive_path = os.path.join(
            self.base_path, "ToolDelta_Game_Texts.tar.gz")
        download_file_singlethreaded(packets_url, archive_path)
        self.extract_data_archive(archive_path)

    def load_data(self) -> Dict[str, str]:
        """加载数据

        Returns:
            Dict[str, str]: 数据
        """
        try:
            all_values: Dict[str, str] = {}
            for file_path in glob(
                os.path.join(self.base_path, "src", "**", "*.py"), recursive=True
            ):
                module_name: str = os.path.basename(
                    file_path).replace(".py", "")
                spec = (
                    util.spec_from_file_location(
                        module_name, file_path)
                )
                if isinstance(spec, type(None)) or isinstance(spec.loader, type(None)):
                    Print.print_war(
                        f"Failed to load module from {file_path}")
                    continue
                module = (
                    util.module_from_spec(spec)
                )
                spec.loader.exec_module(module)
                for var_name in dir(module):
                    if not var_name.startswith("__"):
                        var_value = getattr(module, var_name)
                        if isinstance(var_value, dict):
                            all_values.update(var_value)
            return all_values
        except Exception:
            return {}

    def extract_data_archive(self, zip_path: str) -> bool:
        """解压数据归档

        Args:
            zip_path (str): 压缩包路径

        Returns:
            bool: 是否成功
        """
        try:
            with gzip.open(zip_path, "rb") as f_in, tarfile.open(
                fileobj=f_in, mode="r:gz" # type: ignore
            ) as tar:
                tar.extractall(self.base_path)
                with open(
                    os.path.join(self.base_path, "src_tree.json"), "w", encoding="utf-8"
                ) as f:
                    json.dump(tar.getnames(), f)
                return True
        except Exception as err:
            Print.print_war(f"Error extracting data archive: {err}")
            return False



"""客户端启动器框架"""

import os
import threading
import shlex
import time
import random
import subprocess
import platform
from typing import Callable, Optional
import ujson as json
import requests
import tooldelta

from tooldelta import constants
from . import neo_conn
from .cfg import Cfg
from .builtins import Builtins
from .color_print import Print
from .sys_args import sys_args_to_dict
from .packets import Packet_CommandOutput
from .urlmethod import download_file_singlethreaded, get_free_port

Config = Cfg()


class SysStatus:
    """系统状态码

    LOADING: 启动器正在加载
    LAUNCHING: 启动器正在启动
    RUNNING: 启动器正在运行
    NORMAL_EXIT: 正常退出
    FB_LAUNCH_EXC: FastBuilder 启动异常
    CRASHED_EXIT: 启动器崩溃退出
    NEED_RESTART: 需要重启
    """
    LOADING = 100
    LAUNCHING = 101
    RUNNING = 102
    NORMAL_EXIT = 103
    FB_LAUNCH_EXC = 104
    CRASHED_EXIT = 105
    NEED_RESTART = 106
    launch_type = "None"


class StandardFrame:
    """提供了标准的启动器框架, 作为 ToolDelta 和游戏交互的接口"""
    launch_type = "Original"

    def __init__(self, serverNumber: int, password: str, fbToken: str, auth_server_url: str) -> None:
        """初始化启动器框架

        Args:
            serverNumber (int): 服务器号
            password (str): 服务器密码
            fbToken (str): 验证服务器Token
            auth_server_url (str): 验证服务器地址
        """
        self.serverNumber = serverNumber
        self.serverPassword = password
        self.fbToken = fbToken
        self.auth_server = auth_server_url
        self.system_type = platform.uname().system
        self.inject_events: list = []
        self.packet_handler: Optional[Callable] = lambda pckType, pck: None
        self.need_listen_packets: set[int] = {9, 63, 79}
        self._launcher_listener: Callable
        self.exit_event = threading.Event()
        self.status: int = SysStatus.LOADING

    def add_listen_packets(self, *pcks: int) -> None:
        """添加需要监听的数据包"""
        for i in pcks:
            self.need_listen_packets.add(i)

    def launch(self) -> None:
        """启动器启动

        Raises:
            SystemError: 无法启动此启动器
        """
        raise NotImplementedError

    def listen_launched(self, cb: Callable) -> None:
        """设置监听启动器启动事件"""
        self._launcher_listener = cb

    def get_players_and_uuids(self) -> None:
        """获取玩家名和UUID"""
        raise NotImplementedError

    def get_bot_name(self) -> str:
        """获取机器人名字"""
        raise NotImplementedError

    def update_status(self, new_status: int) -> None:
        """更新启动器状态

        Args:
            new_status (int): 新的状态码
        """
        self.status = new_status
        if new_status == SysStatus.NORMAL_EXIT:
            tooldelta.safe_jump(out_task=True)
            self.exit_event.set()  # 设置事件，触发等待结束
        if new_status == SysStatus.CRASHED_EXIT:
            tooldelta.safe_jump(out_task=False)
            self.exit_event.set()

    def sendcmd(self, cmd: str, waitForResp: bool = False, timeout: int | float = 30) -> Optional[Packet_CommandOutput]:
        """以玩家身份发送命令

        Args:
            cmd (str): 命令
            waitForResp (bool, optional): 是否等待结果
            timeout (int | float, optional): 超时时间

        Raises:
            NotImplementedError: 未实现此方法

        Returns:
            Optional[Packet_CommandOutput]: 返回命令结果
        """
        raise NotImplementedError

    def sendwscmd(self, cmd: str, waitForResp: bool = False, timeout: int | float = 30) -> Optional[Packet_CommandOutput]:
        """以ws身份发送命令

        Args:
            cmd (str): 命令
            waitForResp (bool, optional): 是否等待结果
            timeout (int | float, optional): 超时时间

        Raises:
            NotImplementedError: 未实现此方法

        Returns:
            Optional[Packet_CommandOutput]: 返回命令结果
        """
        raise NotImplementedError

    def sendwocmd(self, cmd: str) -> None:
        """以wo身份发送命令

        Args:
            cmd (str): 命令

        Raises:
            NotImplementedError: 未实现此方法
        """
        raise NotImplementedError

    def sendPacket(self, pckID: int, pck: str) -> None:
        """发送数据包

        Args:
            pckID (int): 数据包ID
            pck (str): 数据包内容

        Raises:
            NotImplementedError: 未实现此方法
        """
        raise NotImplementedError

    sendPacketJson = sendPacket

    def is_op(self, player: str) -> bool:
        """检查玩家是否为OP

        Args:
            player (str): 玩家名

        Raises:
            NotImplementedError: 未实现此方法

        Returns:
            bool: 是否为OP
        """
        raise NotImplementedError


class FrameNeOmg(StandardFrame):
    """使用 NeOmega 框架连接到游戏"""
    launch_type = "NeOmega"

    def __init__(self, serverNumber: int, password: str, fbToken: str, auth_server: str) -> None:
        """初始化 NeOmega 框架

        Args:
            serverNumber (int): 服务器号
            password (str): 服务器密码
            fbToken (str): 验证服务器Token
            auth_server (str): 验证服务器地址
        """
        super().__init__(serverNumber, password, fbToken, auth_server)
        neo_conn.load_lib()
        self.status = SysStatus.LOADING
        self.launch_event = threading.Event()
        self.omega: Optional[neo_conn.ThreadOmega] = None
        self.neomg_proc = None
        self.download_libs()
        self.status = SysStatus.LAUNCHING
        self.secret_exit_key = ""

    def set_omega(self, openat_port: int) -> None:
        """设置 Omega 连接

        Args:
            openat_port (int): 端口号

        Raises:
            SystemExit: 系统退出
        """

        retries = 0
        while retries <= 10:
            try:
                self.omega = neo_conn.ThreadOmega(
                    connect_type=neo_conn.ConnectType.Remote,
                    address=f"tcp://localhost:{openat_port}",
                    accountOption=neo_conn.AccountOptions(
                        AuthServer=self.auth_server,
                        UserToken=self.fbToken,
                        ServerCode=str(self.serverNumber),
                        ServerPassword=str(self.serverPassword),
                    ),
                )
                retries = 0
                break
            except Exception as err:
                time.sleep(5)
                retries += 1
                Print.print_war(f"OMEGA 连接失败, 重连: 第 {retries} 次: {err}")
                if retries > 5:
                    Print.print_err("最大重试次数已达到")
                    raise SystemExit from err

    def start_neomega_proc(self) -> int:
        """启动 NeOmega 进程

        Returns:
            int: 端口号
        """
        free_port = get_free_port(24016)
        access_point_file = (
            f"neomega_{platform.uname().system.lower()}_access_point_{self.sys_machine}"
        )
        if "TERMUX_VERSION" in os.environ:
            access_point_file = f"neomega_android_access_point_{self.sys_machine}"
        if platform.system() == "Windows":
            access_point_file += ".exe"
        py_file_path = os.path.join(
            os.getcwd(), "tooldelta", "neo_libs", access_point_file
        )
        if platform.uname().system.lower() == "linux":
            os.system("chmod +x " + shlex.quote(py_file_path))
        # 只需要+x即可
        Print.print_inf(f"DEBUG: 将使用端口 {free_port}")
        self.neomg_proc = subprocess.Popen(
            [
                py_file_path,
                "-server",
                str(self.serverNumber),
                "-T",
                self.fbToken,
                "-access-point-addr",
                f"tcp://localhost:{free_port}",
                "-server-password",
                str(self.serverPassword),
                "-auth-server",
                self.auth_server,
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        return free_port

    def msg_show(self) -> None:
        """显示来自 NeOmega 的信息"""
        def _msg_show_thread() -> None:
            """显示来自NeOmega的信息"""
            if self.neomg_proc is None or self.neomg_proc.stdout is None:
                raise ValueError("NEOMG 进程未启动")
            while True:
                msg_orig = self.neomg_proc.stdout.readline().decode("utf-8").strip("\n")
                if msg_orig in ("", "SIGNAL: exit"):
                    with Print.lock:
                        Print.print_with_info(
                            "ToolDelta: NEOMG 进程已结束", "§b NOMG ")
                    self.update_status(SysStatus.NORMAL_EXIT)
                    return
                if "[neOmega 接入点]: 就绪" in msg_orig:
                    self.launch_event.set()
                elif f"STATUS CODE: {self.secret_exit_key}" in msg_orig:
                    with Print.lock:
                        Print.print_with_info("§a机器人已退出", "§b NOMG ")
                    continue
                with Print.lock:
                    Print.print_with_info(msg_orig, "§b NOMG ")

        Builtins.createThread(_msg_show_thread, usage="显示来自NeOmega的信息")

    def make_secret_key(self) -> None:
        """生成退出密钥"""
        self.secret_exit_key = hex(random.randint(10000, 99999))

    def launch(self) -> SystemExit | Exception | SystemError:
        """启动 NeOmega 进程

        returns:
            SystemExit: 正常退出
            Exception: 异常退出
            SystemError: 未知的退出状态
        """
        self.status = SysStatus.LAUNCHING
        openat_port = self.start_neomega_proc()
        self.msg_show()
        self.launch_event.wait()
        self.make_secret_key()
        self.set_omega(openat_port)
        self.update_status(SysStatus.RUNNING)
        Print.print_suc("已开启接入点进程")
        if self.omega is None:
            self.update_status(SysStatus.CRASHED_EXIT)
            return SystemExit("接入点进程未启动")
        pcks = [
            self.omega.get_packet_id_to_name_mapping(i)
            for i in self.need_listen_packets
        ]
        self.omega.listen_packets(pcks, self.packet_handler_parent)
        self._launcher_listener()
        Print.print_suc("接入点已就绪!")
        self.exit_event.wait()  # 等待事件的触发
        if self.status == SysStatus.NORMAL_EXIT:
            return SystemExit("正常退出.")
        if self.status == SysStatus.CRASHED_EXIT:
            return Exception("NeOmega 已崩溃")
        return SystemError("未知的退出状态")

    def download_libs(self) -> None:
        """根据系统架构和平台下载所需的库。"""
        cfgs = Config.get_cfg("ToolDelta基本配置.json", constants.LAUNCH_CFG_STD)
        is_mir: bool = cfgs["是否使用github镜像"]
        if is_mir:
            mirror_src = "https://tdload.tblstudio.cn/" + \
                "https://raw.githubusercontent.com/ToolDelta/ToolDelta/main/"
            depen_url = "https://tdload.tblstudio.cn/" + \
                "https://raw.githubusercontent.com/ToolDelta/DependencyLibrary/main/"
        else:
            mirror_src = "https://raw.githubusercontent.com/ToolDelta/ToolDelta/main/"
            depen_url = "https://raw.githubusercontent.com/ToolDelta/DependencyLibrary/main/"
        try:
            require_depen = json.loads(
                requests.get(
                    f"{mirror_src}require_files.json", timeout=5
                ).text
            )
        except Exception as err:
            Print.print_err(f"获取依赖库表出现问题: {err}")
            self.update_status(SysStatus.CRASHED_EXIT)
            return
        self.sys_machine = platform.machine().lower()
        if self.sys_machine == "x86_64":
            self.sys_machine = "amd64"
        elif self.sys_machine == "aarch64":
            self.sys_machine = "arm64"
        if "TERMUX_VERSION" in os.environ:
            sys_info_fmt: str = f"Android:{self.sys_machine.lower()}"
        else:
            sys_info_fmt: str = f"{platform.uname().system}:{self.sys_machine.lower()}"
        source_dict: list[str] = require_depen[sys_info_fmt]
        commit_remote = requests.get(
            f"{depen_url}commit", timeout=5
        ).text
        commit_file_path = os.path.join(
            os.getcwd(), "tooldelta", "neo_libs", "commit")
        replace_file = False
        if os.path.isfile(commit_file_path):
            with open(commit_file_path, "r", encoding="utf-8") as f:
                commit_local = f.read()
            if commit_local != commit_remote:
                Print.print_war("依赖库版本过期, 将重新下载")
                replace_file = True
        else:
            replace_file = True
        for v in source_dict:
            pathdir = os.path.join(os.getcwd(), "tooldelta", "neo_libs", v)
            url = depen_url + v
            if not os.path.isfile(pathdir) or replace_file:
                Print.print_with_info(f"正在下载依赖库 {pathdir} ...", "§a 下载 §r")
                try:
                    download_file_singlethreaded(url, pathdir)
                except Exception as err:
                    Print.print_err(f"下载依赖库出现问题: {err}")
                    self.update_status(SysStatus.CRASHED_EXIT)
                    return
        if replace_file:
            # 写入commit_remote，文字写入
            with open(commit_file_path, "w", encoding="utf-8") as f:
                f.write(commit_remote)
            Print.print_suc("已完成依赖更新！")

    def get_players_and_uuids(self):
        players_uuid = {}
        if self.omega is None:
            raise ValueError("未连接到接入点")
        for i in self.omega.get_all_online_players():
            if i is not None:
                players_uuid[i.name] = i.uuid
            else:
                raise ValueError("未能获取玩家名和UUID")
        return players_uuid

    def get_bot_name(self) -> str:
        """获取机器人名字

        Returns:
            str: 机器人名字
        """
        if self.omega is None:
            raise ValueError("未连接到接入点")
        return self.omega.get_bot_name()

    def packet_handler_parent(self, pkt_type: str, pkt: dict) -> None:
        """数据包处理器

        Args:
            pkt_type (str): 数据包类型
            pkt (dict): 数据包内容

        Raises:
            ValueError: 未连接到接入点
        """
        if self.omega is None or self.packet_handler is None:
            raise ValueError("未连接到接入点")
        packetType = self.omega.get_packet_name_to_id_mapping(pkt_type)
        self.packet_handler(packetType, pkt)

    def sendcmd(self, cmd: str, waitForResp: bool = False, timeout: float = 30) -> Optional[Packet_CommandOutput]:
        """以玩家身份发送命令

        Args:
            cmd (str): 命令
            waitForResp (bool, optional): 是否等待结果
            timeout (int | float, optional): 超时时间

        Raises:
            NotImplementedError: 未实现此方法

        Returns:
            Optional[Packet_CommandOutput]: 返回命令结果
        """
        if self.omega is None:
            raise ValueError("未连接到接入点")
        if waitForResp:
            res = self.omega.send_player_command_need_response(
                cmd, timeout)
            if res is None:
                raise TimeoutError("指令超时")
            return res
        self.omega.send_player_command_omit_response(cmd)
        return

    def sendwscmd(self, cmd: str, waitForResp: bool = False, timeout: float = 30) -> Optional[Packet_CommandOutput]:
        """以玩家身份发送命令

        Args:
            cmd (str): 命令
            waitForResp (bool, optional): 是否等待结果
            timeout (int | float, optional): 超时时间

        Raises:
            NotImplementedError: 未实现此方法

        Returns:
            Optional[Packet_CommandOutput]: 返回命令结果
        """
        if self.omega is None:
            raise ValueError("未连接到接入点")
        if waitForResp:
            res = self.omega.send_websocket_command_need_response(
                cmd, timeout)
            if res is None:
                raise TimeoutError("指令超时")
            return res
        self.omega.send_websocket_command_omit_response(cmd)
        return

    def sendwocmd(self, cmd: str) -> None:
        """以wo身份发送命令

        Args:
            cmd (str): 命令

        Raises:
            NotImplementedError: 未实现此方法
        """
        if self.omega is None:
            raise ValueError("未连接到接入点")
        self.omega.send_settings_command(cmd)

    def sendPacket(self, pckID: int, pck: str) -> None:
        """发送数据包

        Args:
            pckID (int): 数据包ID
            pck (str): 数据包内容

        Raises:
            NotImplementedError: 未实现此方法
        """
        if self.omega is None:
            raise ValueError("未连接到接入点")
        self.omega.send_game_packet_in_json_as_is(pckID, pck)

    def is_op(self, player: str) -> bool:
        """检查玩家是否为OP

        Args:
            player (str): 玩家名

        Raises:
            NotImplementedError: 未实现此方法

        Returns:
            bool: 是否为OP
        """
        if self.omega is None:
            raise ValueError("未连接到接入点")
        player_obj = self.omega.get_player_by_name(player)
        if player_obj is None or player_obj.command_permission_level is None:
            raise ValueError("未能获取玩家对象")
        return player_obj.command_permission_level > 2

    sendPacketJson = sendPacket


class FrameNeOmgRemote(FrameNeOmg):
    """远程启动器框架(使用 NeOmega 框架的Remote连接)

    Args:
        FrameNeOmg (FrameNeOmg): FrameNeOmg 框架
    """
    launch_type = "NeOmega Remote"

    def launch(self) -> SystemExit | Exception | SystemError:
        """启动远程启动器框架

        Raises:
            AssertionError: 端口号错误

        Returns:
            SystemExit | Exception | SystemError: 退出状态
        """
        try:
            openat_port = int(sys_args_to_dict().get(
                "access-point-port") or "24020")
            if openat_port not in range(65536):
                raise AssertionError
        except (ValueError, AssertionError):
            Print.print_err("启动参数 -access-point-port 错误: 不是1~65535的整数")
        if openat_port == 0:
            Print.print_war(
                "未用启动参数指定链接neOmega接入点开放端口, 尝试使用默认端口 24015"
            )
            Print.print_inf("可使用启动参数 -access-point-port 端口 以指定接入点端口.")
            openat_port = 24015
            return SystemExit("未指定端口号")
        Print.print_inf(f"将从端口 {openat_port} 连接至接入点(等待接入中).")
        self.set_omega(openat_port)
        Print.print_suc("已连接上接入点进程.")
        if self.omega is None:
            self.update_status(SysStatus.CRASHED_EXIT)
            return SystemExit("接入点进程未启动")
        pcks = [
            self.omega.get_packet_id_to_name_mapping(i)
            for i in self.need_listen_packets
        ]
        self.omega.listen_packets(pcks, self.packet_handler_parent)
        self._launcher_listener()
        Print.print_suc("接入点已就绪")
        self.exit_event.wait()
        self.update_status(SysStatus.NORMAL_EXIT)
        if self.status == SysStatus.NORMAL_EXIT:
            return SystemExit("正常退出.")
        if self.status == SysStatus.CRASHED_EXIT:
            return Exception("接入点已崩溃")
        return SystemError("未知的退出状态")



"""ToolDelta 启动器"""
import os
import time
import traceback
from .builtins import tmpjson_save_thread
from .urlmethod import check_update
from .frame import Frame, GameCtrl
from .color_print import Print
from .plugin_load.PluginGroup import plugin_group
from .plugin_load.injected_plugin import movent

frame = Frame()


def start_tool_delta() -> None:
    """启动ToolDelta"""
    plugin_group.set_frame(frame)
    try:
        frame.welcome()
        check_update()
        frame.basic_operation()
        frame.loadConfiguration()
        game_control = GameCtrl(frame)
        frame.set_game_control(game_control)
        frame.set_plugin_group(plugin_group)
        movent.set_frame(frame)
        plugin_group.read_all_plugins()
        frame.plugin_load_finished(plugin_group)
        tmpjson_save_thread()
        frame.launcher.listen_launched(game_control.Inject)
        game_control.set_listen_packets()
        raise frame.launcher.launch()
    except (KeyboardInterrupt, SystemExit):
        pass
    except Exception:
        Print.print_err("ToolDelta 运行过程中出现问题: " + traceback.format_exc())


def safe_jump(*, out_task: bool = True, exit_directly: bool = True) -> None:
    """安全退出

    Args:
        out_task (bool, optional): frame框架系统是否退出
        exit_directly (bool, optional): 是否三秒强制直接退出
    """
    if out_task:
        frame.system_exit()
    frame.safelyExit()
    if exit_directly:
        for _ in range(2, 0, -1):
            Print.print_war(f"{_}秒后强制退出...", end="\r")
            time.sleep(1)
        Print.print_war("0秒后强制退出...", end="\r")
        Print.print_suc("ToolDelta 已退出.")
        os._exit(0)
    Print.print_suc("ToolDelta 已退出.")



"简易函数"
import zipfile
from ..color_print import Print


def unzip_plugin(zip_dir: str, exp_dir: str) -> None:
    """解压插件

    Args:
        zip_dir (str): 压缩文件路径
        exp_dir (str): 解压目录
    """
    try:
        f = zipfile.ZipFile(zip_dir, "r")
        f.extractall(exp_dir)
    except Exception as err:
        Print.print_err(f"zipfile: 解压失败: {err}")
        raise EOFError("解压失败") from err



"插件加载主模块"

from typing import Any


def NON_FUNC(*_) -> None:
    "空函数"
    return None


class PluginSkip(EOFError):
    "跳过插件加载"


class NotValidPluginError(AssertionError):
    "仅加载插件时引发, 不是合法插件的报错"


class PluginAPINotFoundError(ModuleNotFoundError):
    "插件API未找到错误"

    def __init__(self, name):
        self.name = name


class PluginAPIVersionError(ModuleNotFoundError):
    "插件API版本错误"

    def __init__(self, name, m_ver, n_ver):
        self.name = name
        self.m_ver = m_ver
        self.n_ver = n_ver


class PluginRegData:
    "插件注册数据"

    def __init__(self, name: str, plugin_data: dict | None = None, is_registered=True, is_enabled=True):
        """插件注册数据

        Args:
            name (str): 插件名
            plugin_data (dict | None, optional): 插件数据
            is_registered (bool, optional): 是否已注册
            is_enabled (bool, optional): 是否启用
        """
        if plugin_data is None:
            plugin_data = {}
        self.name: str = name
        if isinstance(plugin_data.get("version"), str):
            self.version: tuple = tuple(
                int(i) for i in plugin_data.get("version", "0.0.0").split(".")
            )
        else:
            self.version = plugin_data.get("version", (0, 0, 0))
        #  以下的方法或许存在危险
        self.author: str = plugin_data.get("author", "unknown")
        self.plugin_type: str = plugin_data.get("plugin-type", "unknown")
        self.description: str = plugin_data.get("description", "")
        self.pre_plugins: dict[str, str] = plugin_data.get("pre-plugins", [])
        self.plugin_id = plugin_data.get("plugin-id", "???")
        self.is_registered = is_registered
        if plugin_data.get("enabled") is not None:
            self.is_enabled = plugin_data["enabled"]
        else:
            self.is_enabled = is_enabled

    def dump(self) -> dict[str, Any]:
        "转储数据"
        return {
            "author": self.author,
            "version": ".".join([str(i) for i in self.version]),
            "plugin-type": self.plugin_type,
            "description": self.description,
            "pre-plugins": self.pre_plugins,
            "plugin-id": self.plugin_id,
            "enabled": self.is_enabled
        }

    @property
    def version_str(self) -> str:
        """版本字符串

        Returns:
            str: 版本字符串
        """
        return ".".join(str(i) for i in self.version)

    @property
    def plugin_type_str(self) -> str:
        """插件类型字符串

        Returns:
            str: 插件类型字符串
        """
        return {
            "classic": "主类式",
            "injected": "注入式",
            "unknown": "未知类型",
        }.get(self.plugin_type, "未知类型")


def plugin_is_enabled(pname: str) -> bool:
    """插件是否启用

    Args:
        pname (str): 插件名

    Returns:
        bool: 是否启用
    """
    return not pname.endswith("+disabled")
